import win32.windows;
import win32.windef;
import win32.commctrl;

import std.stdio;
import std.string;

import util.Image;

COLORREF revODR(uint col) {
  return cast(COLORREF)((col&0xff00ff00) | ((col&0x00ff0000)>>16) | ((col&0x000000ff)<<16)) ;
}
COLORREF invAlpha(uint col) {
  return cast(COLORREF)((0xff000000-(col&0xff000000)) | (col&0x00ffffff)) ;
}

class BitMap
{

  uint[]   imgbuf ;
  int      w,h ;
  
  this(){
  }

  ~this(){
  }

  int getWidth(){
    return(w) ;
  }

  int getHeight(){
    return(h) ;
  }
  
  int readImage(HINSTANCE hinstance, const wchar* name, const wchar* type){
    HRSRC   hrs = FindResource(hinstance, name, type) ;
    if( hrs!=null ){
      int     size    = SizeofResource(hinstance, hrs);
      HGLOBAL hglobal = LoadResource(hinstance, hrs);
      if( hglobal!=null ){
	ubyte *pdata = cast(ubyte*)(LockResource(hglobal)) ;
	ubyte mdata[] ; mdata = pdata[0..size] ;
	auto image = new Image() ;
	image.open(mdata, "png") ;
	w = image.getWidth()  ;
	h = image.getHeight() ;
	imgbuf = new uint[w*h] ;
	for( int i=0 ; i<h ; i++ ){
	  uint[] img = image.readline(1,true) ;
	  imgbuf[w*i .. w*i+w]=img[0..$] ;
	  delete img ;
	}
	image.close() ;
	FreeResource(hglobal);
	return(0) ;
      }
    }
    return(-1) ;
  }

  int setImage(uint[] img, int width, int height, bool invalpha=false, bool revodr=false){
    if( width*height != img.length ) return(-1) ;
    imgbuf = new uint[width*height] ;
    if( invalpha==false && revodr==false ){
      imgbuf[] = img[] ;
    }else{
      foreach(i, pix ; img ){
	uint col=pix ;
	if(   revodr==true ) col=revODR(col) ;
	if( invalpha==true ) col=invAlpha(col) ;
	imgbuf[i]=col ;
      }
    }
    w=width ;
    h=height ;
    return(0) ;
  }

  HBITMAP makeHBITMAP(uint bgcol){
    HBITMAP    hbitmap ;
    uint       *bmpp ;
    BITMAPINFO bi ;
    
    bi.bmiHeader.biSize             = BITMAPINFOHEADER.sizeof;
    bi.bmiHeader.biWidth            =  w ;
    bi.bmiHeader.biHeight           = -h ;
    bi.bmiHeader.biPlanes           = 1;
    bi.bmiHeader.biBitCount         = 32;
    bi.bmiHeader.biCompression      = BI_RGB;
    bi.bmiHeader.biSizeImage        = 0;
    bi.bmiHeader.biXPelsPerMeter    = 0;
    bi.bmiHeader.biYPelsPerMeter    = 0;
    bi.bmiHeader.biClrUsed          = 1;
    bi.bmiHeader.biClrImportant     = 0;
    hbitmap = CreateDIBSection( null, &bi, DIB_RGB_COLORS, cast(void**)(&bmpp), null, 0 );
    
    for( int y=0 ; y<h ; y++ ){
      bmpp[w*y .. w*y+w]=imgbuf[w*y .. w*y+w] ;
      for( int x=0 ; x<w ; x++ ){
	int br = ((bgcol & 0x00ff0000)>>16) ;
	int bg = ((bgcol & 0x0000ff00)>>8 ) ;
	int bb = ((bgcol & 0x000000ff)    ) ;
	int fa = (bmpp[w*y+x] & 0xff000000)>>24 ;
	int fr = (((bmpp[w*y+x] & 0x000000ff)    ) * fa + br*(255-fa))/255 ;
	int fg = (((bmpp[w*y+x] & 0x0000ff00)>>8 ) * fa + bg*(255-fa))/255 ;
	int fb = (((bmpp[w*y+x] & 0x00ff0000)>>16) * fa + bb*(255-fa))/255 ;
	bmpp[w*y+x] = 0xff000000 | (fr &0xff) | (fg &0xff)<<8 | (fb &0xff)<<16 ;
      }
    }
    return(hbitmap) ;
  }

  int bitblt( uint *bgp, int bgw, int bgh, int dstx, int dsty ){
    for( int y=0 ; y<h ; y++ ){
      for( int x=0 ; x<w ; x++ ){
	int px=dstx+x ;
	int py=dsty+y ;
	if( px>=bgw || py>=bgh || px<0 || py<0 ) continue ;
	uint bgcol = bgp[bgw*py+px] ;
	int br = ((bgcol & 0x000000ff)    ) ;
	int bg = ((bgcol & 0x0000ff00)>>8 ) ;
	int bb = ((bgcol & 0x00ff0000)>>16) ;
	int fa = (imgbuf[w*y+x] & 0xff000000)>>24 ;
	int fr = (((imgbuf[w*y+x] & 0x000000ff)    ) * fa + br*(255-fa))/255 ;
	int fg = (((imgbuf[w*y+x] & 0x0000ff00)>>8 ) * fa + bg*(255-fa))/255 ;
	int fb = (((imgbuf[w*y+x] & 0x00ff0000)>>16) * fa + bb*(255-fa))/255 ;
	bgp[bgw*py+px] = 0xff000000 | (fr &0xff) | (fg &0xff)<<8 | (fb &0xff)<<16 ;
      }
    }
    return(0) ;
  }
    
}  
